#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "constants.h"
#include "diskio.h"
#include "pff.h"

/* The font matrix stored in the PROGMEM zone */
extern PGM_P led_chars[] PROGMEM;

uint8_t leds[6]; /* 6 bytes for the display */
//const char my_string[] = "Proiect PM 2010";
//#define STR_LEN 15

char my_string[MAX_MSG_SIZE + 1];
uint32_t my_string_len;
uint8_t str_idx, letter_idx;
uint32_t main_idx;
uint8_t current_letter[7];

/* FatFS declarations here */
FATFS	Fs;
DIR	Dir;
FILINFO	Fno;
WORD	rb;

void io_init(void);
void timer_init(void);
void load_string_from_file(void);
void render(void);
void clear(void);
void write_to_leds(uint8_t, uint8_t);
void get_letter_code(uint8_t);
uint8_t spi_transfer(uint8_t);

void get_letter_code(uint8_t byte) {

	PGM_P p;
	memcpy_P(&p, &led_chars[byte], sizeof(PGM_P));
	memcpy_P(current_letter, p, FONT_WIDTH);

}

void write_to_leds(uint8_t msg, uint8_t color) {

	uint8_t i;
	uint32_t result = 0, result2 = 0;
	uint32_t tmp_val = color;
	for(i = 0 ; i < 8 ; i++) {
		if(msg & 0x01)
			result |= tmp_val;
		msg >>= 1;
		tmp_val <<= 3;
	}

	result2 = result;
	leds[5] = ~(result & 0xFF);
	result >>= 8;
	leds[4] = ~(result & 0xFF);
	result >>= 8;
	leds[3] = ~(result & 0xFF);

	leds[2] = ~(result2 & 0xFF);
	result2 >>= 8;
	leds[1] = ~(result2 & 0xFF);
	result2 >>= 8;
	leds[0] = ~(result2 & 0xFF);

}

/* This function is called when Timer0 compare matches OCR0 */
ISR( TIMER0_COMP_vect ) {

	if (main_idx >= IMG_SIZE) 
		main_idx = 0;

	if(letter_idx == FONT_WIDTH) {
		str_idx++;
		if(str_idx == my_string_len) str_idx = 0;
		get_letter_code(my_string[str_idx]);
		letter_idx = 0;
	}

//	write_to_leds(current_letter[letter_idx], 2);
	write_to_leds(current_letter[letter_idx], letter_idx + 1);
	render();
	main_idx++;
	letter_idx++;

}

void io_init() {

    DDRB = 0xFF;

    SPCR =  (0<<SPIE) | // We don't want interrupts
	    (1<<SPE)  | // We do want the SPI enabled
	    (1<<DORD) | // We want the data to be shifted out LSB
	    (1<<MSTR) | // We want the atmega to be a master
	    (0<<CPOL) | // We want the leading edge to be rising
	    (0<<CPHA) | // We want the leading edge to be sample
	    (0<<SPR1) | (0<<SPR0); // sets the clock speed

    SPSR =  (0<<SPIF) | // SPI interrupt flag
	    (0<<WCOL) | // Write collision flag
	    (1<<SPI2X); // Doubles the speed of the SPI clock

}

void timer_init() {
/*
	The frequency of the interrupt overflow is determined by the 
	prescaler and overflow value.
	freq = clock_frequency / ( 2 * prescaler * overflow_val)
	where prescaler can be 1, 8, 64, 256, or 1024
	clock_freq is 16MHz
	and overflow_val is 8bit

	The overflow value is placed in OCR0, the prescale is set in TCCR0
	So for example:
	A good POV frequency is around 400Hz, though we will use 200Hz for this test
	desired freq = 200Hz
	clock freq = 16MHz
	16MHz / (200Hz * 256) = 160
	since 160 is less than 256 (largest 8 bit number)
	OCR0 = 160 and the prescale is 256
*/
	TCCR0 |= _BV(WGM01) | _BV(CS02); /* CTC mode, prescaler = 256, overflow_val = 160 */
	OCR0 = 160;
	TIMSK |= _BV(OCIE0); /* Interrupt on compare for timer 0 */
}

uint8_t spi_transfer(uint8_t byte) {

	SPDR = byte;
	while(!(SPSR & _BV(SPIF)));
	return SPDR;

}

inline void wait_instr(int instr) {
	while(instr > 0) instr --;
}

void clear() {

	int i;
	for(i = 0 ; i < 6; i++)
		spi_transfer(0xFF);
	PORTB |= _BV(LED_CS);
	PORTB &= ~_BV(LED_CS);
}

void render() {

	int i;
	for(i = 0 ; i < 6; i++) 
		spi_transfer(leds[5 - i]);
	PORTB |= _BV(LED_CS);
	PORTB &= ~_BV(LED_CS);
}

void load_string_from_file() {

	FRESULT res;
	uint8_t buf;
	
	if((res = pf_open(Fno.fname)) == FR_OK) {

		// Read string size
		do {
		    pf_read(&buf, 1, &rb);
		    if(buf >= '0' && buf <= '9') {
				my_string_len *= 10;
				my_string_len += (buf - '0');
		    }
		} while(buf >= '0' && buf <= '9');
		
		pf_read(my_string, my_string_len, &rb);
	
	}

}

int main(void) {

	int i;

	sei();
	io_init();
	for(i = 0 ; i < 6 ; i++)
		leds[i] = 0xFF;

	/* For debugging purposes when programming via USB bootloader */
	DDRD |= _BV(PD7);
	
	_delay_ms(1000);
	
	/* Load string from file */
	if((rb = pf_mount(&Fs)) != FR_OK) {
	
		/* Print out the error code */
		my_string[0] = rb + '0';
		my_string_len = 2;
		clear();
		main_idx = str_idx = letter_idx = 0;
		get_letter_code(my_string[0]);
	
		timer_init();
		while(1);
	
	}
	
	if((rb = pf_opendir(&Dir, "")) != FR_OK) {

		/* Print out the error code */
		my_string[0] = rb + '0';
		my_string_len = 2;
		clear();
		main_idx = str_idx = letter_idx = 0;
		get_letter_code(my_string[0]);
	
		timer_init();
		while(1);

	}
	
	while(!(rb = pf_readdir(&Dir, &Fno)) && Fno.fname[0]) {
		if(!(Fno.fattrib & (AM_DIR | AM_HID)) && !strcmp(Fno.fname, MSG_FILE)) {
			load_string_from_file();
			break;
		}
	}
	
	main_idx = str_idx = letter_idx = 0;
	get_letter_code(my_string[0]);
	
	timer_init();
	while(1);

	return 0;

}
